import os
import json
import torch
from typing import Union
from .utils import fullname, import_from_string


class Dropout(torch.nn.Module):
    '''
    Dropout layer
    Param dropout: Sets a dropout value for dense layer
    '''
    def __init__(self, dropout: float = 0.2) -> None:
        super().__init__()
        self.dropout = dropout
        self.dropout_layer = torch.nn.Dropout(self.dropout)

    def forward(self, features: torch.Tensor):
        return self.dropout_layer(features)

    def save(self, output_path):
        with open(os.path.join(output_path, 'config.json'), 'w') as f_out:
            json.dump({'dropout': self.dropout}, f_out)

    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as f_in:
            config = json.load(f_in)

        model = Dropout(**config)
        return model


class Dense(torch.nn.Module):
    '''
    One-layer feedforward neural network
    '''
    def __init__(
        self,
        input_features: int,
        output_features: int,
        bias: bool = True,
        activation_function = torch.nn.Sigmoid(),
        init_weights: Union[torch.Tensor, None] = None,
        init_bias: Union[torch.Tensor, None] = None 
        ) -> None:
        super().__init__()
        self.input_features = input_features
        self.output_features = output_features
        self.bias = bias
        self.activation_function = activation_function
        self.linear = torch.nn.Linear(input_features, output_features, bias)

        if init_weights is not None:
            self.linear.weight = torch.nn.Parameter(init_weights)
        if init_bias is not None:
            self.linear.bias = torch.nn.Parameter(init_bias)

    def forward(self, features: torch.Tensor):
        return self.activation_function(self.linear(features))

    def get_config_dict(self):
        return {
            'input_features': self.input_features,
            'output_features': self.output_features,
            'bias': self.bias,
            'activation_function': fullname(self.activation_function)
        }

    def save(self, output_path, file_name='pytorch_model.bin'):
        with open(os.path.join(output_path, 'config.json'), 'w') as f_out:
            json.dump(self.get_config_dict(), f_out)
        torch.save(self.state_dict(), os.path.join(output_path, file_name))

    def __repr__(self):
        return f'Dense({self.get_config_dict()})'

    @staticmethod
    def load(input_path, file_name='pytorch_model.bin'):
        with open(os.path.join(input_path, 'config.json')) as f_in:
            config = json.load(f_in)
        config['activation_function'] = import_from_string(config['activation_function'])()
        model = Dense(**config)
        model.load_state_dict(torch.load(os.path.join(input_path, file_name), map_location=torch.device('cpu')))



import torch


def cross_entropy_loss_fn(outputs, targets, multi_label=False):
    if multi_label:
        return torch.nn.BCELoss()(outputs, targets)
    else:
        return torch.nn.CrossEntropyLoss()(outputs, targets)



'''
Custom training module
'''
import time
import torch
import numpy as np
from tqdm import tqdm
from .io import save_checkpoint
from .metrics import accuracy_metrics


def loss_fn_multiple_labels(outputs, targets):
    '''
    binary cross entropy loss for multi-labels
    '''
    return torch.nn.BCEWithLogitsLoss()(outputs, targets)


def loss_fn(outputs, targets):
    '''
    cross entropy loss function
    '''
    return torch.nn.CrossEntropyLoss()(outputs, targets)


def optimizer(model, learning_rate):
    optimizer = torch.optim.AdamW(
        params=model.parameters(),
        lr=learning_rate
    )
    return optimizer


def custom_trainer(model, optimizer, train_dataloader, test_dataloader, epochs=5, device='cpu'):
    '''
    custom training module
    '''
    for epoch in range(epochs):
        model.train()
        print("")
        print(f'======== Epoch {epoch+1} / {epochs} ========')
        print(f'Total steps: {len(train_dataloader)} || Training in progress...')
        t0 = time.time()
        total_train_loss, total_train_accuracy = 0, 0
        for _, batch in tqdm(enumerate(train_dataloader)):
            # import pdb; pdb.set_trace();
            model.zero_grad(set_to_none=True)
            ids = batch['input_ids'].squeeze(1).to(device)
            mask = batch['attention_mask'].squeeze(1).to(device)
            type_ids = batch['token_type_ids'].squeeze(1).to(device)
            label = batch['label'].to(device)
            output = model(ids, mask, type_ids)

            loss = loss_fn(output, label)
            total_train_loss += loss.item()

            total_train_accuracy += accuracy_metrics(output.detach().cpu().numpy(), label.detach().cpu().numpy())

            # if step % 5000 == 0:
            #     print(f'Epoch: {epoch}, Loss: {loss.item()}')

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
            optimizer.step()
        avg_train_loss = total_train_loss / len(train_dataloader)
        avg_train_accuracy = total_train_accuracy / len(train_dataloader)
        print(f'''
            Training time: {time.time() - t0} seconds ||
            Training loss: {avg_train_loss} ||
            Training accuracy: {avg_train_accuracy}
            ''')
        print('Evaluation in progress...')
        val_outputs, val_targets, val_loss = validate(model, test_dataloader, device)
        avg_val_loss = np.array(val_loss).mean()
        avg_val_accuracy = accuracy_metrics(val_outputs, val_targets)
        print(f'''
            Validation time: {time.time() - t0} seconds
            Validation loss: {avg_val_loss} ||
            Validation accuracy: {avg_val_accuracy}
            ''')
        save_checkpoint(model, optimizer, epoch, avg_train_loss, avg_train_accuracy, avg_val_loss, avg_val_accuracy)


def validate(model, test_dataloader, device='cpu'):
    '''
    produce validation results
    '''
    model.eval()
    val_targets, val_outputs, val_loss = [], [], []
    with torch.no_grad():
        for _, batch in enumerate(test_dataloader):
            ids = batch['input_ids'].squeeze(1).to(device)
            mask = batch['attention_mask'].squeeze(1).to(device)
            type_ids = batch['token_type_ids'].squeeze(1).to(device)
            label = batch['label'].to(device)
            outputs = model(ids, mask, type_ids)
            val_targets.extend(label.detach().cpu().numpy())
            val_outputs.extend(torch.sigmoid(outputs).detach().cpu().numpy())
            val_loss.extend([loss_fn(outputs, label).detach().cpu().numpy()])
    return val_outputs, val_targets, val_loss



import gc
import torch
import transformers
from pathlib import Path
from tqdm.notebook import tqdm
from typing import Optional, Dict
from dataclasses import dataclass, field
from torch.utils.data import DataLoader
from .metrics import accuracy
from .utils import batch_to_device
from .loss import cross_entropy_loss_fn


@dataclass
class Configs:
    epochs: int = 1
    optimizer_class = torch.optim.AdamW
    optimizer_params: Dict[str, float] = field(default_factory = lambda: ({"lr": 2e-5})) 
    weight_decay: float = 0.01
    scheduler: str = 'WarmupLinear'
    warmup_steps: int = 10000
    multi_label: bool = False
    num_labels: Optional[int] = None
    tune_base_model: bool = True


class Trainer:
    def __init__(self, 
                 model: torch.nn.Module,
                 train_dataloader: DataLoader,
                 test_dataloader: DataLoader,
                 configs: Configs,
                 metrics = accuracy,
                 device: torch.device = torch.device('cpu'),
                 chkpt_dir = Path('../chkpt')
                 ):
        self.model = model
        self.train_dataloader = train_dataloader
        self.test_dataloader = test_dataloader
        self.configs = configs
        self.metrics = metrics
        self.device = device
        self.chkpt_dir = chkpt_dir
        self.set_params()
        self.train_logs = []
        self.val_logs = []
    
    def set_params(self):
        '''
        set parameters to optimize and schedule
        '''
        if not self.configs.tune_base_model:
            for param in self.model.base_model.parameters():
                param.requires_grad = False

        self.optimizer = self.get_optimizer(
            list(self.model.named_parameters()), 
            self.configs.optimizer_class, 
            self.configs.optimizer_params, 
            self.configs.weight_decay)
        
        self.scheduler = self.get_scheduler(
            self.optimizer, 
            self.configs.scheduler, 
            self.configs.warmup_steps, 
            len(self.train_dataloader)*self.configs.epochs)
    
    def train(self):
        epochs = tqdm(range(1, self.configs.epochs+1), leave = True, desc="Training...")
        for epoch in epochs:
            self.model.train()
            epochs.set_description(f"EPOCH {epoch} / {self.configs.epochs} | training...")
            self.train_one_epoch(epoch)
            self.clear()
            
            self.model.eval()
            epochs.set_description(f"EPOCH {epoch} / {self.configs.epochs} | validating...")
            self.validate_one_epoch(epoch)
            self.clear()
                
            self.print_per_epoch(epoch)
            self.save_checkpoint(epoch)
            
    def continue_training(self, chkpt_file):
        '''
        continue training from checkpoint
        '''
        self.model, self.optimizer, _ = self.load_checkpoint(chkpt_file, self.model, self.optimizer)
        self.schedule_cold_start()
        self.train()

    def train_one_epoch(self, epoch):
        batches = tqdm(self.train_dataloader, total=len(self.train_dataloader))
        total_train_loss = total_train_acc = 0
        for features, labels in batches:
            labels = labels.to(self.device)
            features = batch_to_device(features, self.device)
            outputs = self.model(**features)
            loss = cross_entropy_loss_fn(outputs[0], labels, self.configs.multi_label)

            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)
            self.optimizer.step()
            self.scheduler.step()

            total_train_loss += loss.item()
            total_train_acc += self.metrics(
                outputs[0], labels, 
                self.configs.multi_label, 
                self.configs.num_labels, 
                self.device).item()
            
            batches.set_description(f"Train Loss Step: {loss.item():.2f}")
        
        self.logger(
            epoch, 
            total_train_acc/len(self.train_dataloader), 
            total_train_loss/len(self.train_dataloader),
            'train')

    @torch.no_grad()
    def validate_one_epoch(self, epoch):
        '''
        Validate module
        '''
        self.model.eval()
        val_targets = torch.tensor(()).to(self.device)
        val_outputs = torch.tensor(()).to(self.device)
        val_loss = torch.tensor(()).to(self.device)

        batches = tqdm(self.test_dataloader, total=len(self.test_dataloader))
        for features, labels in batches:
            labels = labels.to(self.device)
            features = batch_to_device(features, self.device)
            outputs = self.model(**features)
            val_outputs = torch.cat((val_outputs, outputs[0]), 0)
            val_targets = torch.cat((val_targets, labels), 0)
            loss = cross_entropy_loss_fn(outputs[0], labels, self.configs.multi_label).reshape(1)
            val_loss = torch.cat((val_loss, loss), 0)
            batches.set_description(f"Validation Loss Step: {loss.item():.2f}")
        
        avg_val_loss = val_loss.mean().item()
        avg_val_acc = self.metrics(
            val_outputs, val_targets, 
            self.configs.multi_label, 
            self.configs.num_labels, 
            self.device).item()
        
        self.logger(epoch, avg_val_acc, avg_val_loss, 'test')

    @staticmethod
    def get_optimizer(param_optimizer,
                      optimizer_class = torch.optim.AdamW,
                      optimizer_params: dict = {'lr': 2e-5},
                      weight_decay: float = 0.01
                      ):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_grouped_params = [
            {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
            {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
        ]
        optimizer = optimizer_class(optimizer_grouped_params, **optimizer_params)
        return optimizer

    @staticmethod
    def get_scheduler(optimizer, scheduler: str, warmup_steps: int, t_total: int):
        '''
        Returns the correct learning rate scheduler.
        Available scheduler: constantlr, warmupconstant, warmuplinear, warmupcosine, warmupcosinewithhardrestarts
        '''
        scheduler = scheduler.lower()
        if scheduler == 'constantlr':
            return transformers.get_constant_schedule(optimizer)
        elif scheduler == 'warmupconstant':
            return transformers.get_constant_schedule_with_warmup(
                optimizer, num_warmup_steps=warmup_steps)
        elif scheduler == 'warmuplinear':
            return transformers.get_linear_schedule_with_warmup(
                optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        elif scheduler == 'warmupcosine':
            return transformers.get_cosine_schedule_with_warmup(
                optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        elif scheduler == 'warmupcosinewithhardrestarts':
            return transformers.get_cosine_with_hard_restarts_schedule_with_warmup(
                optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        else:
            raise ValueError(f'Unkown scheduler {scheduler}')

    @staticmethod
    def load_checkpoint(chkpt_dir, model, optimizer: Optional=None):
        chkpt = torch.load(chkpt_dir)
        model.load_state_dict(chkpt['model_state_dict'])
        if optimizer:
            optimizer.load_state_dict(chkpt['optimizer_state_dict'])
        return model, optimizer, chkpt
    
    def save_checkpoint(self, epoch):
        if not self.chkpt_dir.exists():
            self.chkpt_dir.mkdir(parents=True)
        epoch -= 1
        chkpt_path = self.chkpt_dir / ('chkpt' + str(epoch) + '.pt')
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'train_loss': self.train_logs[epoch][f'epoch_{epoch}']['loss'],
            'val_loss': self.val_logs[epoch][f'epoch_{epoch}']['loss'],
            'train_acc': self.train_logs[epoch][f'epoch_{epoch}']['accuracy'],
            'val_acc': self.val_logs[epoch][f'epoch_{epoch}']['accuracy'],
        }, chkpt_path)
        
    def schedule_cold_start(self):
        self.scheduler = self.get_scheduler(
            self.optimizer, 
            self.configs.scheduler, 
            0, 
            len(self.train_dataloader)*self.configs.epochs)

    def print_per_epoch(self, epoch):
        print(f"\n\n{'-'*30}EPOCH {epoch}/{self.configs.epochs}{'-'*30}")
        epoch -= 1 
        train_loss = self.train_logs[epoch][f'epoch_{epoch}']['loss']
        train_acc = self.train_logs[epoch][f'epoch_{epoch}']['accuracy']
        val_loss = self.val_logs[epoch][f'epoch_{epoch}']['loss']
        val_acc = self.val_logs[epoch][f'epoch_{epoch}']['accuracy']
        print(f"Train -> LOSS: {train_loss} | ACCURACY: {train_acc}")
        print(f"Validation -> LOSS: {val_loss} | ACCURACY: {val_acc}\n\n\n")

    def logger(self, epoch, metrics, loss, mode):
        log = {
            f'epoch_{epoch-1}': {
                'loss': loss,
                'accuracy': metrics
            }
        }
        if mode == 'train':
            self.train_logs.append(log)
        else:
            self.val_logs.append(log)

    def clear(self):
        gc.collect()
        torch.cuda.empty_cache()



from typing import Optional, Literal
import torch
from torchmetrics.classification import (
    MultilabelHammingDistance,
    MulticlassHammingDistance,
    MultilabelAccuracy,
    MulticlassAccuracy
)


def hamming_distance(
    outputs,
    targets,
    multi_label: bool = False, 
    num_labels: Optional[int] = None, 
    device: torch.device = torch.device('cpu'),
    average: Literal['micro', 'macro', 'weighted', 'none'] = 'macro'
):
    if not num_labels:
        raise ValueError('num_labels is required')
    if multi_label:
        hamming = MultilabelHammingDistance(num_labels, average=average)
    else:
        hamming = MulticlassHammingDistance(num_labels, average=average)
    hamming.to(device)
    return hamming(outputs, targets)


def accuracy(
    outputs,
    targets,
    num_labels: int,
    multi_label: bool = False,
    device: torch.device = torch.device('cpu'),
    average: Literal['micro', 'macro', 'weighted', 'none'] = 'macro'
):
    if not num_labels:
        raise ValueError('num_labels is required')
    if multi_label:
        acc = MultilabelAccuracy(num_labels, average=average)
    else:
        acc = MulticlassAccuracy(num_labels, average=average)
    acc.to(device)
    return acc(outputs, targets)



'''
bert inheritance class and functions
'''
import torch
from .io import get_pretrained_model


class BertClass(torch.nn.Module):
    '''
    Custom class for bert classification model
    '''
    def __init__(self, num_label, base_model=None):
        super().__init__()
        self.l1 = base_model if base_model else get_pretrained_model()
        self.l2 = torch.nn.Dropout(self.l1.config.hidden_dropout_prob)
        self.l3 = torch.nn.Linear(self.l1.config.hidden_size, num_label)

    def forward(self, input_ids, attention_mask, token_type_ids):
        # import pdb; pdb.set_trace();
        last_hidden_state = self.l1(input_ids, attention_mask, token_type_ids).last_hidden_state
        cls_representation = last_hidden_state[:, 0, :]
        dropout_output = self.l2(cls_representation)
        pooled_output = self.l3(dropout_output)
        return pooled_output


def bert_encoder(content, tokenizer, max_len):
    '''
    encoder function with preset params
    '''
    inputs = tokenizer.encode_plus(
        content,
        add_special_tokens=True,
        truncation=True,
        padding='max_length',
        max_length=max_len,
        return_attention_mask=True,
        return_token_type_ids=True,
        return_tensors='pt'
    )
    return inputs



'''
Inference module
'''
import torch
from .bert import bert_encoder
from .io import get_pretrained_tokenizer


MAX_LEN = 512


class Inference:
    '''
    inference class
    '''
    def __init__(self, model, labels, tokenizer=None, max_len=MAX_LEN):
        self.tokenizer = tokenizer if tokenizer else get_pretrained_tokenizer()
        self.model = model
        self.labels = labels
        self.max_len = max_len

    def predict(self, inp: str):
        '''
        inp: text input
        '''
        self.model.eval()
        enc = bert_encoder(inp, self.tokenizer, self.max_len)
        out = self.model(**enc)[-1].detach().cpu()
        idx = out.argmax().item()
        label = self.labels[idx]
        proba = out.softmax(-1)[idx].item()
        return label, proba

    def __call__(self, inp):
        return self.predict(inp)


def load_model(model_dir, device='cpu'):
    '''
    load full model
    '''
    return torch.load(model_dir, map_location=device)



import importlib
import numpy as np
from torch import Tensor, device


def fullname(o):
    '''
    Give a full name (package_name.class_name) for a class / object in Python.
    Used to load the correct classes from JSON files
    '''
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    else:
        return module + '.' + o.__class__.__name__
    

def import_from_string(dotted_path):
    '''
    Import a dotted module path and return the attribute/class designed by
    the last name in the path. Raise ImportError if the import failed.
    '''
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = f'{dotted_path} doesn\'t look like a module path'
        raise ImportError(msg)
    
    try:
        module = importlib.import_module(dotted_path)
    except Exception:
        module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = f'Module {module_path} does not define a {class_name} attribute/class'
        raise ImportError(msg)
    

def batch_to_device(batch, target_device: device):
    '''
    Send a pytorch batch to a device (CPU/GPU)
    '''
    for key in batch:
        if isinstance(batch[key], Tensor):
            batch[key] = batch[key].squeeze(1).to(target_device)
    return batch


def get_embedding_group(label, cat_lookup, corpus):
    idx_grp = cat_lookup[label]
    corpus_embeddings, question_corpus, answer_corpus = corpus
    corpus_embeddings = corpus_embeddings[idx_grp]
    question_corpus = np.array(question_corpus)[idx_grp].tolist()
    answer_corpus = np.array(answer_corpus)[idx_grp].tolist()
    return corpus_embeddings, question_corpus, answer_corpus



'''
data modules
'''
import torch
from torch.utils.data import Dataset
from .bert import bert_encoder


class CustomDataset(Dataset):
    '''
    Class to construct torch Dataset from dataframe
    '''
    def __init__(self, dataframe, data_field, label_field, tokenizer, max_len, multi_label):
        self.max_len = max_len
        self.data = dataframe
        self.tokenizer = tokenizer
        self.content = self.data[data_field]
        self.label = self.data[label_field]
        self.multi_label = multi_label

    def __len__(self):
        return len(self.content)

    def __getitem__(self, index):
        content = str(self.content[index])
        content = " ".join(content.split())
        encoded_content = self.tokenizer.encode_plus(
            content,
            add_special_tokens=True,
            truncation=True,
            padding='max_length',
            max_length=self.max_len,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='pt'
        )
        features = encoded_content
        if self.multi_label:
            labels = torch.tensor(self.label[index], dtype=torch.float)
        else:
            labels = torch.tensor(self.label[index], dtype=torch.long)

        return features, labels


def create_label_dict(dataframe, label_col):
    labels = dataframe.groupby(label_col).size().sort_values(ascending=False).index.tolist()
    label_dict = dict([(d, i) for i, d in enumerate(labels)])
    return label_dict


def label2id(dataframe, label_col, label_dict, multi_label=False):
    if multi_label:
        dataframe[label_col] = dataframe[label_col].apply(lambda c: [int(k in c) for k in label_dict.keys()])
    else:
        dataframe[label_col] = dataframe[label_col].apply(lambda c: label_dict[c])
    return dataframe



import torch
import transformers
import numpy as np
from .layers import Dropout, Dense
from .utils import get_embedding_group


class AutoModelForMTL(torch.nn.Module):
    '''
    Formation of MTL model, sharing a same base model
    '''
    def __init__(self, base_model_name, num_labels, no_normalize=False, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.base_model = transformers.AutoModel.from_pretrained(base_model_name)
        self.dropout_layer = Dropout()
        self.dense_layer = Dense(self.base_model.config.hidden_size, num_labels, activation_function=torch.nn.Sigmoid())
        self.normalize = not no_normalize

    def forward(self, **kwargs):
        model_output = self.base_model(**kwargs)
        sentence_embedding = self.mean_pooling(model_output, kwargs['attention_mask'])
        if self.normalize:
            sentence_embedding = torch.nn.functional.normalize(sentence_embedding, p=2, dim=1)
        cls_embedding = self.cls_pooling(model_output)
        dropout_output = self.dropout_layer(cls_embedding)
        pooled_output = self.dense_layer(dropout_output)
        return pooled_output, sentence_embedding

    def mean_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def cls_pooling(self, model_output):
        return model_output[0][:, 0]
    
    @staticmethod
    def save_model(model, dir):
        torch.save(model, dir)

    @staticmethod
    def load_model(dir, device=torch.device('cpu')):
        return torch.load(dir, device)


class MTLInference:
    '''
    class for MTL model infering
    '''
    def __init__(self, tokenizer_card, model_card, num_labels=None, pretrained_model=True, device=torch.device('cpu')) -> None:
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(tokenizer_card)
        if pretrained_model:
            self.mtl_model = AutoModelForMTL(model_card, num_labels)
        else:
            self.mtl_model = AutoModelForMTL.load_model(model_card, device)
        self.mtl_model.eval()

    def encode(self, inputs):
        return self.tokenizer(inputs, return_tensors='pt', padding=True)

    def predict(self, query):
        with torch.no_grad():
            encoded = self.encode(query)
            outputs = self.mtl_model(**encoded)
        pred_label = outputs[0]
        query_embedding = outputs[1]
        return pred_label, query_embedding

    def optimal_answer(self, query, cat_lookup, corpus, top_k=1):
        pred_label, query_embedding = self.predict(query)
        corpus_embeddings, question_corpus, answer_corpus = get_embedding_group(pred_label, cat_lookup, corpus)
        sim_scores = torch.mm(query_embedding, corpus_embeddings.transpose(0, 1).cpu().tolist())
        top_idx = np.argpartition(sim_scores, range(-top_k, 0))[-top_k:][::-1]
        score = list(sim_scores[top_idx])
        question_matched = [question_corpus[i] for i in top_idx]
        answer_matched = [answer_corpus[i] for i in top_idx]
        return dict(
            question_matched=question_matched,
            answer_matched=answer_matched,
            score=score
        )



'''
IO utils
'''
import json
from pathlib import Path
import torch
from transformers import BertTokenizer, BertModel


MODEL_NAME = 'bert-base-uncased'
MODEL_DIR = Path('../models/bert')
PRETRAINED_TOKENIZER = MODEL_DIR / 'pretrained/tokenizer-uncased'
PRETRAINED_MODEL = MODEL_DIR / 'pretrained/bert-base-uncased'
FINETUNED_DIR = MODEL_DIR / 'fine-tuned'
FINETUNED_MODEL = FINETUNED_DIR / 'fine-tuned-uncased'
FINETUNED_MODEL_STATE = FINETUNED_DIR / 'model-state-dict'
FINETUNED_OPT_STATE = FINETUNED_DIR / 'opt-state-dict'
CHECKPOINT_DIR = MODEL_DIR / 'checkpoint'
LABEL_DICT = MODEL_DIR / 'fine-tuned/labels-dict.json'


def get_pretrained_tokenizer(model_name=MODEL_NAME, pretrained_tokenizer_dir=PRETRAINED_TOKENIZER):
    '''
    get pretrained tokenizer
    '''
    if pretrained_tokenizer_dir.exists():
        pretrained_tokenizer = BertTokenizer.from_pretrained(pretrained_tokenizer_dir)
    else:
        pretrained_tokenizer = BertTokenizer.from_pretrained(model_name)
        pretrained_tokenizer.save_pretrained(pretrained_tokenizer_dir)
    return pretrained_tokenizer


def get_pretrained_model(model_name=MODEL_NAME, pretrained_model_dir=PRETRAINED_MODEL):
    '''
    get pretrained model
    '''
    if pretrained_model_dir.exists():
        pretrained_model = BertModel.from_pretrained(pretrained_model_dir)
    else:
        pretrained_model = BertModel.from_pretrained(model_name)
        pretrained_model.save_pretrained(pretrained_model_dir)
    return pretrained_model


def save_checkpoint(model, optimizer, epoch, train_loss, train_accuracy, val_loss, val_accuracy):
    '''
    save checkpoint
    '''
    if not CHECKPOINT_DIR.exists():
        CHECKPOINT_DIR.mkdir(parents=True)
    chkpt_path = CHECKPOINT_DIR / ('chkpt' + str(epoch) + '.pt')
    torch.save({
        'epoch': epoch,
        'model_state_dict': model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict(),
        'train_loss': train_loss,
        'val_loss': val_loss,
        'train_acc': train_accuracy,
        'val_acc': val_accuracy,
    }, chkpt_path)


def load_checkpoint(model, optimizer, chkpoint_dir):
    '''
    load checkpoint
    '''
    checkpoint = torch.load(chkpoint_dir)
    model.load_state_dict(checkpoint['model_state_dict'])
    optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    epoch = checkpoint['epoch']
    train_loss = checkpoint['train_loss']
    val_loss = checkpoint['val_loss']
    train_acc = checkpoint['train_acc']
    val_acc = checkpoint['val_acc']
    return model, optimizer, epoch, train_loss, val_loss, train_acc, val_acc


def save_model(model, optimizer):
    '''
    save model para
    '''
    if not FINETUNED_DIR.exists():
        FINETUNED_DIR.mkdir(parents=True)
    torch.save(model.state_dict(), FINETUNED_MODEL_STATE)
    torch.save(optimizer.state_dict(), FINETUNED_OPT_STATE)
    torch.save(model, FINETUNED_MODEL)


def load_model(model_dir=FINETUNED_MODEL, device='cpu'):
    '''
    load the whole model
    '''
    return torch.load(model_dir, map_location=torch.device(device))


def load_model_safe(num_label):
    '''
    preferred and more flexible way to load model
    '''
    from .bert import BertClass
    pretrained_model = get_pretrained_model(PRETRAINED_MODEL, MODEL_NAME)
    model = BertClass(pretrained_model, num_label)
    model.load_state_dict(torch.load(FINETUNED_MODEL_STATE))
    return model


def save_label_dict(label_dict, dict_file=LABEL_DICT):
    '''save label dictionary'''
    with open(dict_file, 'w') as file:
        json.dump(label_dict, file)


def load_label_dict(label_file=LABEL_DICT):
    '''load label dictionary'''
    with open(label_file, 'r') as file:
        return json.load(file)



'''
Performance metrics
'''
import numpy as np
from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    accuracy_score,
    confusion_matrix
)


def transform_outputs(outputs, targets, multi_label=False):
    '''
    transform outputs to suitable format for calculating performance metrics
    '''
    preds = np.array(outputs).argmax(axis=-1)
    if multi_label:
        y = np.array(targets).argmax(axis=1)
    else:
        y = targets
    return y, preds


def accuracy_metrics(outputs, targets, multi_label=False):
    '''
    function to calculate accuracy
    outputs: model outputs
    targets: labels
    '''
    targets, preds = transform_outputs(outputs, targets, multi_label)
    return accuracy_score(targets, preds)


def classification_metrics(outputs, targets, multi_label=False):
    '''
    collection of classification metrics
    '''
    targets, preds = transform_outputs(outputs, targets, multi_label)
    return dict([
        ('n', len(targets)),
        ('accuracy', accuracy_score(targets, preds)),
        ('precision', precision_score(targets, preds, average='macro')),
        ('recall', recall_score(targets, preds, average='macro')),
        ('f1', f1_score(targets, preds, average='macro')),
        ('confusion', confusion_matrix(targets, preds))
    ])
